( cd usr/bin ; rm -rf scons )
( cd usr/bin ; ln -sf /usr/bin/scons-0.97 scons )
( cd usr/bin ; rm -rf sconsign )
( cd usr/bin ; ln -sf /usr/bin/sconsign-0.97 sconsign )
( cd usr/bin ; rm -rf scons-time )
( cd usr/bin ; ln -sf /usr/bin/scons-time-0.97 scons-time )
